import { Router } from "express";
import { JobType, TypedRequest, USER_TYPE } from "../config/types";
import { authMiddleware, organizationMiddleware } from "../middlewares/auth";
import Auth from "../models/Auth";
import Jobs from "../models/Job";
import Application from "../models/Application";

const router = Router();

router.get(
  "/",
  authMiddleware,
  async (req: TypedRequest<{ userId: string }, null>, res) => {
    try {
      const user = await Auth.findOne({ userId: req.query.userId });

      let jobs;

      if (user.userType === USER_TYPE.CANDIDATE) {
        jobs = await Jobs.find()
          .populate({
            path: "applications",
            select: ["candidate", "status", "interview"],
          })
          .populate({ path: "organization" });
      } else {
        jobs = await Jobs.find({ organization: req.query.userId }).populate({
          path: "applications",
          populate: {
            path: "candidate",
            select: [
              "name",
              "contacts",
              "dateOfBirth",
              "profilePicture",
              "resume",
              "resumeData",
              "selfIntro",
            ],
          },
        });
      }

      return res.json({ jobs, success: true });
    } catch (error) {
      return res.json({ error, success: false });
    }
  }
);

router.post(
  "/",
  authMiddleware,
  organizationMiddleware,
  async (req: TypedRequest<{ userId: string }, JobType>, res) => {
    try {
      const newJob = new Jobs({ ...req.body, organization: req.query.userId });
      const job = await newJob.save();
      return res.json({ success: true, job });
    } catch (error) {
      return res.json({ success: false, error });
    }
  }
);

router.put(
  "/",
  authMiddleware,
  organizationMiddleware,
  async (req: TypedRequest<{ userId: string }, JobType>, res) => {
    try {
      const job = await Jobs.findByIdAndUpdate(req.body._id, {
        $set: req.body,
      });
      return res.json({ success: true, job });
    } catch (error) {
      return res.json({ success: false, error });
    }
  }
);

router.delete(
  "/",
  authMiddleware,
  organizationMiddleware,
  async (req: TypedRequest<{ userId: string }, { jobId: string }>, res) => {
    try {
      await Jobs.deleteOne({
        organization: req.query.userId,
        _id: req.body.jobId,
      });
      return res.json({ success: true });
    } catch (error) {
      return res.json({ success: false });
    }
  }
);

router.get(
  "/search",
  authMiddleware,
  (req: TypedRequest<{ key: string }, null>, res) => {
    console.log(req.query.key);
    const pipeline = [
      {
        $search: {
          index: "searchJobs",
          text: {
            query: req.query.key,
            path: "title",
            fuzzy: {},
          },
        },
      },
    ];
    Jobs.aggregate(pipeline)
      .then((data) => res.status(200).json({ data }))
      .catch(() => {
        res.status(200).json({ data: [] });
      });
  }
);

router.delete(
  "/close",
  authMiddleware,
  async (req: TypedRequest<{}, { jobId: string }>, res) => {
    try {
      await Jobs.findByIdAndDelete(req.body.jobId);
      await Application.deleteMany({
        job: req.body.jobId,
      });
      return res.status(200).send("SUCCESS");
    } catch (error: any) {
      return res.status(400).send(error.message);
    }
  }
);

module.exports = router;
