import * as _ from "lodash";
import * as bcrypt from "bcryptjs";
import {
  AuthType,
  ControlsType,
  KeyDetails,
  KeystrokeDataType,
  KeystrokeType,
  USER_TYPE,
} from "../config/types";
import Auth from "../models/Auth";

const ss = require("simple-statistics");

const SD_MULTIPLIER = 2.5;
const THRESHOLD_PERCENT = 65;
const FULL_STANDARD_THRESHOLD = 1;
const STANDARD_NORM_MULTIPLIER = 1000;
const TYPES = ["hold", "flight", "dd", "full"];

const norm = (a: number[]) => Math.sqrt(ss.sum(a.map((v) => v ** 2)));

const euclidean = (d1: number, d2: number, w = 1) =>
  Math.sqrt(w * Math.abs(d2 - d1) ** 2);

const cityblock = (d1: number, d2: number) => Math.abs(d1 - d2);

const cityblockArray = (a1: number[], a2: number[]) =>
  ss.sum(a1.map((v, i) => Math.abs(a1[i] - a2[i])));

const covariance = (
  x: number[],
  y: number[],
  xMean: number,
  yMean: number,
  n: number
) =>
  (1 / (n - 1)) *
  ss.sum(
    Array(n)
      .fill(0)
      .map((v, i) => (x[i] - xMean) * (y[i] - yMean))
  );

export const processKeystrokeData = ({
  keydown,
  keyup,
}: {
  keydown: KeyDetails[];
  keyup: KeyDetails[];
}): KeystrokeDataType => {
  const data: KeystrokeDataType = {
    hold: {
      keys: [],
      times: [],
    },
    flight: {
      keys: [],
      times: [],
    },
    dd: {
      keys: [],
      times: [],
    },
    full: {
      keys: [],
      times: [],
    },
  };

  for (let i = 0; i < keydown.length; i += 1) {
    const { code: downCode, key: downKey, time: downTime } = keydown[i];
    const { code: upCode, key: upKey, time: upTime } = keyup[i];
    const holdTime = upTime - downTime;

    data.full.keys.push(`H:${downKey}`);
    data.full.times.push(holdTime);

    data.hold.keys.push(downKey);
    data.hold.times.push(holdTime);

    if (i < keydown.length - 1) {
      const { key: nextDownKey, time: nextDownTime } = keydown[i + 1];
      const keyString = `${downKey}:${nextDownKey}`;
      const flightTime = nextDownTime - upTime;
      const ddTime = nextDownTime - downTime;

      data.full.keys.push(`F:${keyString}`);
      data.full.times.push(flightTime);

      data.full.keys.push(`DD:${keyString}`);
      data.full.times.push(ddTime);

      data.flight.keys.push(keyString);
      data.flight.times.push(flightTime);

      data.dd.keys.push(keyString);
      data.dd.times.push(ddTime);
    }
  }

  return data;
};

const computeStandardTendencies = (data: KeystrokeType) => {
  const transposedData = _.unzip(data.times);

  const means = data?.means?.map((v, i) => ss.mean(transposedData[i]));
  const sd = data?.sd?.map((v, i) => ss.standardDeviation(transposedData[i]));

  return { means, sd };
};

const computeFilteredTendencies = (
  data: KeystrokeType,
  SDFilterMultiplier = 2
) => {
  const transposedData = _.unzip(data.times).map((row) =>
    row.filter(
      // Each Row is now the columns of original matrix
      // Each element (attempt for a given keystroke) is checked and filtered
      (v: any, i) =>
        euclidean(v, data?.means?.[i]) <
        SDFilterMultiplier * (data?.sd?.[i] || 1)
    )
  );

  const filteredMeans = data?.filteredMeans?.map((v, i) =>
    ss.mean(transposedData[i].length === 0 ? [0] : transposedData[i])
  );
  const filteredSd = data?.filteredSd?.map((v, i) =>
    ss.standardDeviation(
      transposedData[i].length === 0 ? [0] : transposedData[i]
    )
  );

  return { filteredMeans, filteredSd };
};

const computeMahalanobis = (data: KeystrokeType, means: number[]) => {
  const times = data.times as number[][];
  const transTime = _.unzip(times);

  const obsCount = data.times.length;
  const featureCount = times[0]?.length;

  const covMatrix = new Array(featureCount)
    .fill(0)
    .map(() => new Array(featureCount).fill(0));

  Array(featureCount)
    .fill(0)
    .map((v, i) => {
      Array(featureCount)
        .fill(0)
        .map((w, j) => {
          covMatrix[i][j] = covariance(
            transTime[i],
            transTime[j],
            means[i],
            means[j],
            obsCount
          );

          return j;
        });
      return i;
    });

  return { covMatrix };
};

export const computeDataTendencies = (keystrokeData: KeystrokeDataType) => {
  TYPES.map((type) => {
    const key = type as keyof KeystrokeDataType;
    const { means, sd } = computeStandardTendencies(keystrokeData[key]);
    keystrokeData[key].means = means;
    keystrokeData[key].sd = sd;

    const { filteredMeans, filteredSd } = computeFilteredTendencies(
      keystrokeData[key]
    );
    keystrokeData[key].filteredMeans = filteredMeans;
    keystrokeData[key].filteredSd = filteredSd;

    const { covMatrix } = computeMahalanobis(keystrokeData[key], means || []);
    keystrokeData[key].covMatrix = covMatrix;

    return key;
  });

  return keystrokeData;
};

export const createSignupDataFromProcessedData = async (
  email: string,
  passwords: string[],
  userType: USER_TYPE,
  userId: string,
  processedData: KeystrokeDataType[] | any
) => {
  const keystroke = {
    keys: [],
    times: [],
    sums: [],
    means: [],
    sd: [],
    filteredMeans: [],
    filteredSd: [],
    covMatrix: [],
  };

  const salt = await bcrypt.genSalt(10);

  const hashedPassword = await bcrypt.hash(passwords[0], salt);

  let signupData: AuthType = {
    email,
    userType,
    userId,
    password: hashedPassword,
    keystrokeDataTimestamps: [],
    keystrokeData: {
      hold: keystroke,
      flight: keystroke,
      dd: keystroke,
      full: keystroke,
    },
  };

  signupData = processedData.reduce((acc, v, i) => {
    TYPES.map((type) => {
      const key = type as keyof KeystrokeDataType;
      if (i === 0) {
        // Keys of the processed data is checked for equality
        // So the first one is used for the dataset
        acc.keystrokeData[key].keys = v[key].keys;
        const length = v[key].keys.length;

        // Length of the processedData array is the number of attempts
        // Times is an array of arrays of each attempt
        acc.keystrokeData[key].times = Array(v.length).fill([]);

        acc.keystrokeData[key].means = Array(length).fill(0);
        acc.keystrokeData[key].sd = Array(length).fill(0);
        acc.keystrokeData[key].filteredMeans = Array(length).fill(0);
        acc.keystrokeData[key].filteredSd = Array(length).fill(0);
      }
      acc.keystrokeData[key].times[i] = v[key].times;

      return type;
    });

    acc.keystrokeDataTimestamps.push(Date.now());

    return acc;
  }, signupData);

  signupData.keystrokeData = computeDataTendencies(signupData.keystrokeData);

  return signupData;
};

export const findUser = (userId: string) => Auth.findOne({ userId }).exec();

export const addAttemptToKeystrokeData = ({
  userData,
  attemptKeystrokeData,
}: {
  userData: AuthType;
  attemptKeystrokeData: KeystrokeDataType;
}) => {
  TYPES.map((type) => {
    const key = type as keyof KeystrokeDataType;
    userData.keystrokeData[key].times.concat(attemptKeystrokeData[key].times);
    return type;
  });
  userData.keystrokeDataTimestamps.push(Date.now());

  userData.keystrokeData = computeDataTendencies(userData.keystrokeData);

  return userData;
};

const processStandard = (
  userKeystrokeData: KeystrokeDataType,
  attemptKeystrokeData: KeystrokeDataType,
  controls: ControlsType["standard"]
) => {
  const { threshold = THRESHOLD_PERCENT, sd: sdMult = SD_MULTIPLIER } =
    controls;

  const scores: any = {
    distance: {},
    inRange: {},
    inRangePercent: {},
  };

  TYPES.map((type) => {
    const key = type as keyof KeystrokeDataType;
    scores.distance[key] = Array(userKeystrokeData[key].means?.length).fill(0);
    const isInRange = Array(userKeystrokeData[key].means?.length).fill(false);
    let inRangeCount: any = 0;
    let totalCount = 0;
    scores.inRangePercent[key] = 0;
    scores.inRange[key] = false;

    userKeystrokeData[key].means?.map((v, i) => {
      const mean = userKeystrokeData[key].means?.[i] || 0;
      const sd = userKeystrokeData[key].sd?.[i] || 0;
      const attemptTime = attemptKeystrokeData[key].times[i];

      const low = mean - sd * sdMult;
      const high = mean + sd * sdMult;

      const distance = cityblock(mean, attemptTime);

      scores.distance[key][i] = distance;
      isInRange[i] = attemptTime >= low && attemptTime <= high;

      inRangeCount += !!isInRange[i];
      totalCount += 1;
      return v;
    });

    scores.inRangePercent[key] = (inRangeCount / totalCount) * 100;
    if (scores.inRangePercent[key] >= threshold) {
      scores.inRange[key] = true;
    }

    return key;
  });

  return scores;
};

const processFullStandard = (
  userKeystrokeData: KeystrokeDataType,
  attemptKeystrokeData: KeystrokeDataType,
  controls: ControlsType["fullStandard"]
) => {
  const { threshold = FULL_STANDARD_THRESHOLD } = controls;

  const scores: any = {
    normedDistance: {},
    distance: {},
    inRange: {},
  };

  TYPES.map((type) => {
    const key = type as keyof KeystrokeDataType;
    scores.distance[key] = 0;
    scores.inRange[key] = false;

    const distance = cityblockArray(
      attemptKeystrokeData[key].times,
      userKeystrokeData[key].means || []
    );
    scores.distance[key] = distance;

    const normedDistance =
      (distance * STANDARD_NORM_MULTIPLIER) /
      (norm(attemptKeystrokeData[key].times) *
        norm(userKeystrokeData[key].means || []));

    scores.normedDistance[key] = normedDistance;
    if (normedDistance <= threshold) {
      scores.inRange[key] = true;
    }

    return key;
  });

  return scores;
};

export const processAttempt = ({
  userKeystrokeData,
  attemptKeystrokeData,
  controls,
}: {
  userKeystrokeData: KeystrokeDataType;
  attemptKeystrokeData: KeystrokeDataType;
  controls: ControlsType;
}) => {
  const { standard, fullStandard } = controls;

  const result = {
    standard: {
      ...standard,
      ...processStandard(userKeystrokeData, attemptKeystrokeData, standard),
    },
    fullStandard: {
      ...fullStandard,
      ...processFullStandard(
        userKeystrokeData,
        attemptKeystrokeData,
        fullStandard
      ),
    },
    accepted: false,
  };

  const standardCheck = controls.standard.use
    ? result.standard.inRange.full
    : true;

  const fullStandardCheck = controls.fullStandard.use
    ? result.fullStandard.inRange.full
    : true;

  result.accepted = standardCheck || fullStandardCheck;

  return result;
};

export const updatePasswordFromProcessedData = async (
  passwords: string[],
  processedData: KeystrokeDataType[] | any
) => {
  const keystroke = {
    keys: [],
    times: [],
    sums: [],
    means: [],
    sd: [],
    filteredMeans: [],
    filteredSd: [],
    covMatrix: [],
  };

  const salt = await bcrypt.genSalt(10);

  const hashedPassword = await bcrypt.hash(passwords[0], salt);

  let newPasswordData: Partial<AuthType> = {
    password: hashedPassword,
    keystrokeDataTimestamps: [],
    keystrokeData: {
      hold: keystroke,
      flight: keystroke,
      dd: keystroke,
      full: keystroke,
    },
  };

  newPasswordData = processedData.reduce((acc, v, i) => {
    TYPES.map((type) => {
      const key = type as keyof KeystrokeDataType;
      if (i === 0) {
        // Keys of the processed data is checked for equality
        // So the first one is used for the dataset
        acc.keystrokeData[key].keys = v[key].keys;
        const length = v[key].keys.length;

        // Length of the processedData array is the number of attempts
        // Times is an array of arrays of each attempt
        acc.keystrokeData[key].times = Array(v.length).fill([]);

        acc.keystrokeData[key].means = Array(length).fill(0);
        acc.keystrokeData[key].sd = Array(length).fill(0);
        acc.keystrokeData[key].filteredMeans = Array(length).fill(0);
        acc.keystrokeData[key].filteredSd = Array(length).fill(0);
      }
      acc.keystrokeData[key].times[i] = v[key].times;

      return type;
    });

    acc.keystrokeDataTimestamps.push(Date.now());

    return acc;
  }, newPasswordData);

  newPasswordData.keystrokeData = computeDataTendencies(
    newPasswordData.keystrokeData
  );

  return newPasswordData;
};
